# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

from rucio.common import exception
from rucio.common.constants import DEFAULT_VO
from rucio.core import credential
from rucio.core.rse import get_rse_id
from rucio.db.sqla.constants import DatabaseOperationType
from rucio.db.sqla.session import db_session
from rucio.gateway import permission

if TYPE_CHECKING:
    from rucio.common.constants import RSE_BASE_SUPPORTED_PROTOCOL_OPERATIONS_LITERAL, SUPPORTED_SIGN_URL_SERVICES_LITERAL


def get_signed_url(
    account: str,
    rse: str,
    service: 'SUPPORTED_SIGN_URL_SERVICES_LITERAL',
    operation: 'RSE_BASE_SUPPORTED_PROTOCOL_OPERATIONS_LITERAL',
    url: str,
    lifetime: int,
    vo: str = DEFAULT_VO,
) -> str:
    """
    Get a signed URL for a particular service and operation.

    The signed URL will be valid for 1 hour.

    :param account: Account identifier as a string.
    :param rse: The name of the RSE to which the URL points.
    :param service: The service to authorise, currently only 'gsc'.
    :param operation: The operation to sign, either 'read', 'write', or 'delete'.
    :param url: The URL to sign.
    :param lifetime: Lifetime in seconds.
    :param vo: The vo to act on.

    :returns: Signed URL as a variable-length string.
    """

    kwargs = {'account': account}

    with db_session(DatabaseOperationType.READ) as session:
        auth_result = permission.has_permission(issuer=account, vo=vo, action='get_signed_url', kwargs=kwargs, session=session)
        if not auth_result.allowed:
            raise exception.AccessDenied('Account %s can not get signed URL for rse=%s, service=%s, operation=%s, url=%s, lifetime=%s. %s' %
                                         (account, rse, service, operation, url, lifetime, auth_result.message))

        # look up RSE ID for name
        rse_id = get_rse_id(rse, vo=vo, session=session)

    return credential.get_signed_url(rse_id, service, operation, url, lifetime)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Any, Optional

import rucio.gateway.permission
from rucio.common.constants import DEFAULT_VO
from rucio.common.exception import AccessDenied, InvalidAccountType
from rucio.common.schema import validate_schema
from rucio.common.types import InternalAccount
from rucio.common.utils import gateway_update_return_dict
from rucio.core import account as account_core
from rucio.core.rse import get_rse_id
from rucio.db.sqla.constants import AccountType, DatabaseOperationType
from rucio.db.sqla.session import db_session

if TYPE_CHECKING:
    from collections.abc import Iterator

    from rucio.common.types import AccountAttributesDict, IdentityDict, UsageDict
    from rucio.db.sqla.models import Account


def add_account(
    account: str,
    type_: str,
    email: str,
    issuer: str,
    vo: str = DEFAULT_VO,
) -> None:
    """
    Creates an account with the provided account name, contact information, etc.

    :param account: The account name.
    :param type_: The account type
    :param email: The Email address associated with the account.

    :param issuer: The issuer account_core.
    :param vo: The VO to act on.

    """

    validate_schema(name='account', obj=account, vo=vo)

    kwargs = {'account': account, 'type': type_}
    if type_.upper() not in AccountType._member_names_:
        raise InvalidAccountType(f"{type_} is an invalid account type. Choose from {AccountType._member_names_}")

    with db_session(DatabaseOperationType.WRITE) as session:
        auth_result = rucio.gateway.permission.has_permission(issuer=issuer, vo=vo, action='add_account', kwargs=kwargs, session=session)
        if not auth_result.allowed:
            raise AccessDenied('Account %s can not add account. %s' % (issuer, auth_result.message))

        internal_account = InternalAccount(account, vo=vo)

        account_core.add_account(internal_account, AccountType[type_.upper()], email, session=session)


def del_account(
    account: str,
    issuer: str,
    vo: str = DEFAULT_VO,
) -> None:
    """
    Disables an account with the provided account name.

    :param account: The account name.
    :param issuer: The issuer account.
    :param vo: The VO to act on.

    """
    kwargs = {'account': account}
    with db_session(DatabaseOperationType.WRITE) as session:
        auth_result = rucio.gateway.permission.has_permission(issuer=issuer, vo=vo, action='del_account', kwargs=kwargs, session=session)
        if not auth_result.allowed:
            raise AccessDenied('Account %s can not delete account. %s' % (issuer, auth_result.message))

        internal_account = InternalAccount(account, vo=vo)

        account_core.del_account(internal_account, session=session)


def get_account_info(
    account: str,
    vo: str = DEFAULT_VO,
) -> "Account":
    """
    Returns the info like the statistics information associated to an account_core.

    :param account: The account name.
    :returns: A list with all account information.
    :param vo: The VO to act on.

    """

    internal_account = InternalAccount(account, vo=vo)

    with db_session(DatabaseOperationType.READ) as session:
        acc = account_core.get_account(internal_account, session=session)
        acc.account = acc.account.external
        return acc


def update_account(
    account: str,
    key: str,
    value: Any,
    issuer: str = 'root',
    vo: str = DEFAULT_VO,
) -> None:
    """ Update a property of an account_core.

    :param account: Name of the account_core.
    :param key: Account property like status.
    :param value: Property value.
    :param issuer: The issuer account
    :param vo: The VO to act on.

    """
    validate_schema(name='account', obj=account, vo=vo)
    kwargs = {}
    with db_session(DatabaseOperationType.WRITE) as session:
        auth_result = rucio.gateway.permission.has_permission(issuer=issuer, vo=vo, action='update_account', kwargs=kwargs, session=session)
        if not auth_result.allowed:
            raise AccessDenied('Account %s can not change %s  of the account. %s' % (issuer, key, auth_result.message))

        internal_account = InternalAccount(account, vo=vo)

        return account_core.update_account(internal_account, key, value, session=session)


def list_accounts(filter_: Optional[dict[str, Any]] = None, vo: str = DEFAULT_VO) -> 'Iterator[dict[str, Any]]':
    """
    Lists all the Rucio account names.

    REST API: http://<host>:<port>/rucio/accounts

    :param filter_: Dictionary of attributes by which the input data should be filtered
    :param vo: The VO to act on.

    :returns: List of all accounts.
    """
    # If filter is empty, create a new dict to avoid overwriting the function's default
    filter_ = filter_ or {}

    if 'account' in filter_:
        filter_['account'] = InternalAccount(filter_['account'], vo=vo)
    else:
        filter_['account'] = InternalAccount(account='*', vo=vo)

    with db_session(DatabaseOperationType.READ) as session:
        for result in account_core.list_accounts(filter_=filter_, session=session):
            yield gateway_update_return_dict(result, session=session)


def account_exists(
    account: str,
    vo: str = DEFAULT_VO,
) -> bool:
    """
    Checks to see if account exists. This procedure does not check it's status.

    :param account: Name of the account.
    :param vo: The VO to act on.

    :returns: True if found, otherwise false.
    """

    internal_account = InternalAccount(account, vo=vo)

    with db_session(DatabaseOperationType.READ) as session:
        return account_core.account_exists(internal_account, session=session)


def list_identities(
    account: str,
    vo: str = DEFAULT_VO,
) -> list["IdentityDict"]:
    """
    List all identities on an account_core.

    :param account: The account name.
    :param vo: The VO to act on.

    """

    internal_account = InternalAccount(account, vo=vo)

    with db_session(DatabaseOperationType.READ) as session:
        return account_core.list_identities(internal_account, session=session)


def list_account_attributes(
    account: str,
    vo: str = DEFAULT_VO,
) -> list["AccountAttributesDict"]:
    """
    Returns all the attributes for the given account.

    :param account: The account name.
    :param vo: The VO to act on

    """

    internal_account = InternalAccount(account, vo=vo)

    with db_session(DatabaseOperationType.READ) as session:
        return account_core.list_account_attributes(internal_account, session=session)


def add_account_attribute(
    key: str,
    value: Any,
    account: str,
    issuer: str,
    vo: str = DEFAULT_VO,
) -> None:
    """
    Add an attribute to an account.

    :param key: attribute key.
    :param value: attribute value.
    :param account: The account name.
    :param issuer: The issuer account.
    :param vo: The VO to act on.

    """
    validate_schema(name='account_attribute', obj=key, vo=vo)
    validate_schema(name='account_attribute', obj=value, vo=vo)

    kwargs = {'account': account, 'key': key, 'value': value}
    with db_session(DatabaseOperationType.WRITE) as session:
        auth_result = rucio.gateway.permission.has_permission(issuer=issuer, vo=vo, action='add_attribute', kwargs=kwargs, session=session)
        if not auth_result.allowed:
            raise AccessDenied('Account %s can not add attributes. %s' % (issuer, auth_result.message))

        internal_account = InternalAccount(account, vo=vo)

        account_core.add_account_attribute(internal_account, key, value, session=session)


def del_account_attribute(
    key: str,
    account: str,
    issuer: str,
    vo: str = DEFAULT_VO,
) -> None:
    """
    Delete an attribute to an account.

    :param key: attribute key.
    :param account: The account name.
    :param issuer: The issuer account.
    :param vo: The VO to act on.

    """
    kwargs = {'account': account, 'key': key}
    with db_session(DatabaseOperationType.WRITE) as session:
        auth_result = rucio.gateway.permission.has_permission(issuer=issuer, vo=vo, action='del_attribute', kwargs=kwargs, session=session)
        if not auth_result.allowed:
            raise AccessDenied('Account %s can not delete attribute. %s' % (issuer, auth_result.message))

        internal_account = InternalAccount(account, vo=vo)

        account_core.del_account_attribute(internal_account, key, session=session)


def get_usage(
    rse: str,
    account: str,
    issuer: str,
    vo: str = DEFAULT_VO,
) -> "UsageDict":
    """
    Returns current values of the specified counter, or raises CounterNotFound if the counter does not exist.

    :param rse:              The RSE.
    :param account:          The account name.
    :param issuer:           The issuer account.
    :param vo:               The VO to act on.

    :returns:                A dictionary with total and bytes.
    """
    with db_session(DatabaseOperationType.READ) as session:
        rse_id = get_rse_id(rse=rse, vo=vo, session=session)
        internal_account = InternalAccount(account, vo=vo)

        return account_core.get_usage(rse_id, internal_account, session=session)


def get_usage_history(
    rse: str,
    account: str,
    issuer: str,
    vo: str = DEFAULT_VO,
) -> list["UsageDict"]:
    """
    Returns historical values of the specified counter, or raises CounterNotFound if the counter does not exist.

    :param rse:              The RSE.
    :param account:          The account name.
    :param issuer:           The issuer account.
    :param vo:               The VO to act on.

    :returns:                A dictionary with total and bytes.
    """
    with db_session(DatabaseOperationType.READ) as session:
        rse_id = get_rse_id(rse=rse, vo=vo, session=session)
        internal_account = InternalAccount(account, vo=vo)

        return account_core.get_usage_history(rse_id, internal_account, session=session)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from json import dumps
from typing import TYPE_CHECKING, Any, Literal, Optional

from requests.status_codes import codes

from rucio.client.baseclient import BaseClient, choice
from rucio.common.constants import HTTPMethod
from rucio.common.utils import build_url

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping


class DiracClient(BaseClient):
    """
    Client for the DIRAC integration layer.

    This client wraps the REST calls used by the ``RucioFileCatalog`` plugin in DIRAC.
    Only `add_files` is currently provided and it behaves like any other ``BaseClient``
    method by handling authentication tokens and host selection automatically.
    """

    DIRAC_BASEURL = 'dirac'

    def add_files(
            self,
            lfns: "Iterable[Mapping[str, Any]]",
            ignore_availability: bool = False,
            parents_metadata: Optional["Mapping[str, Mapping[str, Any]]"] = None
    ) -> Literal[True]:
        """
        Register files and create missing parent structures.

        For each entry in ``lfns`` the method:

        * Creates the file and its replica on the specified RSE.
        * If the containing dataset does not exist, it is created with a replication
          rule using the RSE expression ``ANY=true``. This places the dataset on any
          RSE advertising the ``ANY`` attribute.
        * Creates all ancestor containers when needed.
        * Attaches metadata from ``parents_metadata`` to those parents.

        Parameters
        ----------
        lfns
            Iterable of dictionaries describing the files. Each dictionary must contain:

            * **``lfn``**  full logical file name with scope
            * **``rse``**  destination RSE name
            * **``bytes``**  file size in bytes
            * **``adler32``**  Adler‑32 checksum

            Optional keys include ``guid`` ``pfn`` and ``meta``.

        ignore_availability
            When ``True``, the availability status of RSEs is ignored and blocked RSEs are
            still accepted. Defaults to ``False`` which rejects blocked RSEs.
        parents_metadata
            Mapping of parent logical path names to metadata {'lpn': {key : value}}.
            Entries are only applied when new datasets or containers are created.
            Defaults to None.

        Returns
        -------
        Literal[True]
            When the server confirms the creation.

        Raises
        ------
        RucioException
            Raised when the HTTP request is not successful.

        Examples
        --------
        ??? Example

            Register a file using the DIRAC naming style. Dirac's scope extraction is
            required to be set for this to work:

            ```python
            >>> from rucio.client.diracclient import DiracClient
            >>> from rucio.common.utils import generate_uuid

            >>> dc = DiracClient()
            >>> lfn = f"/belle/mock/cont_{generate_uuid()}/dataset_{generate_uuid()}/file_{generate_uuid()}"
            >>> files = [{
            ...     "lfn": lfn,
            ...     "rse": "XRD1",
            ...     "bytes": 1,
            ...     "adler32": "0cc737eb",
            ...     'guid': generate_uuid()
            ... }]

            >>> dc.add_files(files)
            True
            ```
        """
        path = '/'.join([self.DIRAC_BASEURL, 'addfiles'])
        url = build_url(choice(self.list_hosts), path=path)

        r = self._send_request(
            url,
            method=HTTPMethod.POST,
            data=dumps({'lfns': lfns, 'ignore_availability': ignore_availability, 'parents_metadata': parents_metadata})
        )

        if r.status_code == codes.created:
            return True
        else:
            exc_cls, exc_msg = self._get_exception(headers=r.headers, status_code=r.status_code, data=r.content)
            raise exc_cls(exc_msg)



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import TYPE_CHECKING, Optional

import click
from rich.text import Text
from tabulate import tabulate

from rucio.cli.utils import JSONType
from rucio.client.richclient import CLITheme, generate_table, get_cli_config, print_output
from rucio.common.constants import OPENDATA_DID_STATE_LITERAL_LIST
from rucio.common.utils import extract_scope

if TYPE_CHECKING:
    from click import Context

    from rucio.common.constants import OPENDATA_DID_STATE_LITERAL

cli_config = get_cli_config()


def is_valid_json(s: str) -> bool:
    try:
        json.loads(s)
        return True
    except json.JSONDecodeError:
        return False


@click.group()
def opendata() -> None:
    """Manage Opendata resources"""


@opendata.group(name="did")
def opendata_did() -> None:
    """Manage Opendata DIDs"""


@opendata_did.command("list")
@click.option("--state", type=click.Choice(OPENDATA_DID_STATE_LITERAL_LIST, case_sensitive=False), required=False,
              help="Filter on Opendata state")
@click.option("--public", required=False, is_flag=True, default=False,
              help="Perform request against the public endpoint")
@click.option("--short", is_flag=True, default=False, help="Dump the list of Opendata DIDs")
@click.pass_context
def list_opendata_dids(ctx: "Context", state: Optional["OPENDATA_DID_STATE_LITERAL"], public: bool,
                       short: bool) -> None:
    """
    List Opendata DIDs, optionally filtered by state and public/private access
    """

    client = ctx.obj.client
    spinner = ctx.obj.spinner

    dids_list = client.list_opendata_dids(state=state, public=public)

    table_data = []

    if cli_config == 'rich':
        spinner.update(status='Fetching Opendata DIDs')
        spinner.start()

    for did in dids_list["dids"]:
        if cli_config == 'rich':
            table_data.append([f"{did['scope']}:{did['name']}",
                               Text(did['state'], style=CLITheme.OPENDATA_DID_STATE.get(did['state'], 'default'))])
        else:
            table_data.append([f"{did['scope']}:{did['name']}", did['state']])

    if short:
        for did, _ in table_data:
            print(did)
    else:
        if cli_config == 'rich':
            table = generate_table(table_data, headers=['SCOPE:NAME', '[STATE]'], col_alignments=['left', 'left'])
            spinner.stop()
            print_output(table, console=ctx.obj.console, no_pager=ctx.obj.no_pager)
        else:
            print(tabulate(table_data, tablefmt="psql", headers=['SCOPE:NAME', '[STATE]']))


@opendata_did.command("add")
@click.argument("did")
@click.pass_context
def add_opendata_did(ctx: "Context", did: str) -> None:
    """
    Adds an existing DID to the Opendata catalog
    """

    client = ctx.obj.client
    scope, name = extract_scope(did)
    client.add_opendata_did(scope=scope, name=name)


@opendata_did.command("remove")
@click.argument("did")
@click.pass_context
def remove_opendata_did(ctx: "Context", did: str) -> None:
    """
    Removes an existing Opendata DID from the Opendata catalog
    """

    client = ctx.obj.client
    scope, name = extract_scope(did)
    client.remove_opendata_did(scope=scope, name=name)


@opendata_did.command("show")
@click.argument("did")
@click.option("--meta", required=False, is_flag=True, default=False, help="Print only the opendata metadata")
@click.option("--files", required=False, is_flag=True, default=False,
              help="Print the files associated with the opendata DID")
@click.option("--public", required=False, is_flag=True, default=False,
              help="Perform request against the public endpoint")
@click.pass_context
def get_opendata_did(ctx: "Context", did: str, files: bool, meta: bool, public: bool) -> None:
    """
    Get information about an Opendata DID, optionally including files and metadata.
    """

    client = ctx.obj.client
    spinner = ctx.obj.spinner
    console = ctx.obj.console

    scope, name = extract_scope(did)
    info = client.get_opendata_did(scope=scope, name=name, public=public,
                                   include_files=files, include_metadata=meta,
                                   include_doi=True)

    output = []
    if cli_config == 'rich':
        spinner.update(status='Fetching Opendata DID stats')
        spinner.start()
        keyword_styles = {**CLITheme.BOOLEAN, **CLITheme.OPENDATA_DID_STATE}

        table_data = [(k, Text(str(v), style=keyword_styles.get(str(v), 'default'))) for (k, v) in
                      sorted(info.items())]
        table = generate_table(table_data, row_styles=['none'], col_alignments=['left', 'left'])
        output.append(table)
    else:
        table = [(k + ':', str(v)) for (k, v) in sorted(info.items())]
        print(tabulate(table, tablefmt='plain', disable_numparse=True))

    if cli_config == 'rich':
        spinner.stop()
        print_output(*output, console=console, no_pager=ctx.obj.no_pager)


@opendata_did.command("update")
@click.argument("did")
@click.option("--meta", type=JSONType(), required=False, help="Opendata JSON")
@click.option("--state", type=click.Choice(OPENDATA_DID_STATE_LITERAL_LIST, case_sensitive=False), required=False,
              help="State of the Opendata DID")
@click.option("--doi", required=False,
              help="Digital Object Identifier (DOI) for the Opendata DID (e.g., 10.1234/foo.bar)")
@click.pass_context
def update_opendata_did(ctx: "Context", did: str, meta: Optional[str],
                        state: Optional["OPENDATA_DID_STATE_LITERAL"],
                        doi: Optional[str]) -> None:
    """
    Update an existing Opendata DID in the Opendata catalog.
    """

    if not any([meta, state, doi]):
        raise ValueError("At least one of --meta, --state, or --doi must be provided.")

    client = ctx.obj.client
    spinner = ctx.obj.spinner
    console = ctx.obj.console

    scope, name = extract_scope(did)
    info = client.update_opendata_did(scope=scope, name=name, meta=meta, state=state, doi=doi)

    if cli_config == 'rich':
        spinner.update(status='Fetching Opendata DID stats')
        spinner.start()
        keyword_styles = {**CLITheme.BOOLEAN, **CLITheme.OPENDATA_DID_STATE}

        table_data = [(k, Text(str(v), style=keyword_styles.get(str(v), 'default'))) for (k, v) in
                      sorted(info.items())]
        table = generate_table(table_data, row_styles=['none'], col_alignments=['left', 'left'])
        spinner.stop()
        print_output(table, console=console, no_pager=ctx.obj.no_pager)
    else:
        table = [(k + ':', str(v)) for (k, v) in sorted(info.items())]
        print(tabulate(table, tablefmt='plain', disable_numparse=True))



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from json import dumps

from flask import Flask, Response, request

from rucio.common.constants import HTTPMethod
from rucio.common.exception import AccessDenied, InvalidObject, LifetimeExceptionDuplicate, LifetimeExceptionNotFound, UnsupportedOperation
from rucio.common.utils import APIEncoder
from rucio.gateway.lifetime_exception import add_exception, list_exceptions, update_exception
from rucio.web.rest.flaskapi.authenticated_bp import AuthenticatedBlueprint
from rucio.web.rest.flaskapi.v1.common import ErrorHandlingMethodView, check_accept_header_wrapper_flask, generate_http_error_flask, json_parameters, param_get, response_headers, try_stream


class LifetimeException(ErrorHandlingMethodView):
    """ REST APIs for Lifetime Model exception. """

    @check_accept_header_wrapper_flask(['application/x-json-stream'])
    def get(self):
        """
        ---
        summary: List Exceptions
        description: "Retrieves all exceptions."
        tags:
            - Lifetime Exceptions
        responses:
          200:
            description: "OK"
            content:
              application/x-json-stream:
                schema:
                  description: "One exception per line."
                  type: array
                  items:
                    description: "A lifetime exception"
                    type: object
                    properties:
                      id:
                        description: "The id of the lifetime exception."
                        type: string
                      scope:
                        description: "The scope associated with the lifetime exception."
                        type: string
                      name:
                        description: "The name of the lifetime exception."
                        type: string
                      did_type:
                        description: "The type of the DID."
                        type: string
                        enum: ['F', 'D', 'C', 'A', 'X', 'Y', 'Z']
                      account:
                        description: "The account associated with the lifetime exception."
                        type: string
                      pattern:
                        description: "The pattern of the lifetime exception."
                        type: string
                      comments:
                        description: "The comments of the lifetime exception."
                        type: string
                      state:
                        description: "The state of the lifetime exception."
                        type: string
                        enum: ['A', 'R', 'W']
                      created_at:
                        description: "The datetime the lifetime exception was created."
                        type: string
                      expires_at:
                        description: "The datetime the lifetime exception expires."
                        type: string
          401:
            description: "Invalid Auth Token"
          404:
            description: "Lifetime exception not found"
          406:
            description: "Not acceptable"
        """
        try:
            def generate(vo):
                for exception in list_exceptions(vo=vo):
                    yield dumps(exception, cls=APIEncoder) + '\n'

            return try_stream(generate(vo=request.environ['vo']))
        except LifetimeExceptionNotFound as error:
            return generate_http_error_flask(404, error)

    def post(self):
        """
        ---
        summary: Create Exception
        description: "Creates a Lifetime Exception."
        tags:
            - Lifetime Exceptions
        requestBody:
          content:
            application/json:
              schema:
                type: object
                properties:
                  dids:
                    description: "List of DIDs associated with the lifetime exception."
                    type: array
                    items:
                      description: "A DID"
                      type: object
                      properties:
                        name:
                          description: "The name of the DID."
                          type: string
                  pattern:
                    description: "The pattern of the lifetime exception."
                    type: string
                  comments:
                    description: "The comment for the lifetime exception."
                    type: string
                  expires_at:
                    description: "The expiration date for the lifetime exception."
                    type: string
        responses:
          201:
            description: "OK"
            content:
              application/json:
                schema:
                  description: "The exception id."
                  type: string
          401:
            description: "Invalid Auth Token"
          400:
            description: "Cannot decode json parameter list."
          409:
            description: "Lifetime exception already exists."
        """
        parameters = json_parameters()
        try:
            exception_id = add_exception(
                dids=param_get(parameters, 'dids', default=[]),
                account=request.environ['issuer'],
                vo=request.environ['vo'],
                pattern=param_get(parameters, 'pattern', default=None),
                comments=param_get(parameters, 'comments', default=None),
                expires_at=param_get(parameters, 'expires_at', default=None),
            )
        except InvalidObject as error:
            return generate_http_error_flask(400, error)
        except AccessDenied as error:
            return generate_http_error_flask(401, error)
        except LifetimeExceptionDuplicate as error:
            return generate_http_error_flask(409, error)

        return Response(dumps(exception_id), status=201, content_type="application/json")


class LifetimeExceptionId(ErrorHandlingMethodView):
    """ REST APIs for Lifetime Model exception. """

    @check_accept_header_wrapper_flask(['application/x-json-stream'])
    def get(self, exception_id):
        """
        ---
        summary: Get Exception
        description: "Get a single Lifetime Exception."
        tags:
            - Lifetime Exceptions
        parameters:
        - name: exception_id
          in: path
          description: "The id of the lifetime exception."
          schema:
            type: string
          style: simple
        responses:
          200:
            description: "OK"
            content:
              application/x-json-stream:
                schema:
                  description: "List of lifetime exceptions associated with the id."
                  type: array
                  items:
                    description: "A lifetime exception"
                    type: object
                    properties:
                      id:
                        description: "The id of the lifetime exception."
                        type: string
                      scope:
                        description: "The scope associated with the lifetime exception."
                        type: string
                      name:
                        description: "The name of the lifetime exception."
                        type: string
                      did_type:
                        description: "The type of the DID."
                        type: string
                        enum: ['F', 'D', 'C', 'A', 'X', 'Y', 'Z']
                      account:
                        description: "The account associated with the lifetime exception."
                        type: string
                      pattern:
                        description: "The pattern of the lifetime exception."
                        type: string
                      comments:
                        description: "The comments of the lifetime exception."
                        type: string
                      state:
                        description: "The state of the lifetime exception."
                        type: string
                        enum: ['A', 'R', 'W']
                      created_at:
                        description: "The datetime the lifetime exception was created."
                        type: string
                      expires_at:
                        description: "The datetime the lifetime exception expires."
                        type: string
          401:
            description: "Invalid Auth Token"
          404:
            description: "Lifetime exception not found"
          406:
            description: "Not acceptable"
        """
        try:
            def generate(vo):
                for exception in list_exceptions(exception_id, vo=vo):
                    yield dumps(exception, cls=APIEncoder) + '\n'

            return try_stream(generate(vo=request.environ['vo']))
        except LifetimeExceptionNotFound as error:
            return generate_http_error_flask(404, error)

    def put(self, exception_id):
        """
        ---
        summary: Approve/Reject exception
        description: "Approve/Reject a Lifetime Exception."
        tags:
            - Lifetime Exceptions
        parameters:
        - name: exception_id
          in: path
          description: "The id of the Lifetime Exception."
          schema:
            type: string
          style: simple
        requestBody:
          content:
            application/json:
              schema:
                type: object
                properties:
                  state:
                    description: "The new state for the Lifetime Exception."
                    type: string
                    enum: ['A', 'R']
        responses:
          201:
            description: "OK"
            content:
              application/json:
                schema:
                  type: string
                  enum: ['Created']
          401:
            description: "Invalid Auth Token"
          404:
            description: "Lifetime Exception not found"
          400:
            description: "Cannot decode json parameter list."
        """
        parameters = json_parameters()
        state = param_get(parameters, 'state', default=None)

        try:
            update_exception(exception_id=exception_id, state=state, issuer=request.environ['issuer'], vo=request.environ['vo'])
        except UnsupportedOperation as error:
            return generate_http_error_flask(400, error)
        except AccessDenied as error:
            return generate_http_error_flask(401, error)
        except LifetimeExceptionNotFound as error:
            return generate_http_error_flask(404, error)

        return 'Created', 201


def blueprint() -> AuthenticatedBlueprint:
    bp = AuthenticatedBlueprint('lifetime_exceptions', __name__, url_prefix='/lifetime_exceptions')

    lifetime_exception_view = LifetimeException.as_view('lifetime_exception')
    bp.add_url_rule('/', view_func=lifetime_exception_view, methods=[HTTPMethod.GET.value, HTTPMethod.POST.value])
    lifetime_exception_id_view = LifetimeExceptionId.as_view('lifetime_exception_id')
    bp.add_url_rule('/<exception_id>', view_func=lifetime_exception_id_view, methods=[HTTPMethod.GET.value, HTTPMethod.PUT.value])

    bp.after_request(response_headers)
    return bp


def make_doc():
    """ Only used for sphinx documentation """
    doc_app = Flask(__name__)
    doc_app.register_blueprint(blueprint())
    return doc_app




'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '38.5.0',
    'branch_nick': 'release-38-LTS',
    'revision_id': '430fd3dd8f4dc5103e1e932c9515421e1c515c3e',
    'revno': 14073
}



# Copyright European Organization for Nuclear Research (CERN) since 2012
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING, Any, Optional

from rucio.common import exception
from rucio.common.constants import DEFAULT_VO
from rucio.common.types import InternalAccount, InternalScope
from rucio.common.utils import gateway_update_return_dict
from rucio.core import lifetime_exception
from rucio.db.sqla.constants import DatabaseOperationType
from rucio.db.sqla.session import db_session
from rucio.gateway import permission

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from rucio.db.sqla.constants import LifetimeExceptionsState


def list_exceptions(
    exception_id: Optional[str] = None,
    states: Optional["Iterable[LifetimeExceptionsState]"] = None,
    vo: str = DEFAULT_VO,
) -> 'Iterator[dict[str, Any]]':
    """
    List exceptions to Lifetime Model.

    :param id:         The id of the exception
    :param states:     The states to filter
    :param vo:         The VO to act on
    """

    with db_session(DatabaseOperationType.READ) as session:
        exceptions = lifetime_exception.list_exceptions(exception_id=exception_id, states=states, session=session)
        for e in exceptions:
            if vo == e['scope'].vo:
                yield gateway_update_return_dict(e, session=session)


def add_exception(
    dids: "Iterable[dict[str, Any]]",
    account: str,
    pattern: str,
    comments: str,
    expires_at: str,
    vo: str = DEFAULT_VO,
) -> dict[str, Any]:
    """
    Add exceptions to Lifetime Model.

    :param dids:        The list of DIDs
    :param account:     The account of the requester.
    :param pattern:     The account.
    :param comments:    The comments associated to the exception.
    :param expires_at:  The expiration date of the exception.
    :param vo:          The VO to act on.

    returns:            The id of the exception.
    """

    internal_account = InternalAccount(account, vo=vo)
    for did in dids:
        did['scope'] = InternalScope(did['scope'], vo=vo)

    with db_session(DatabaseOperationType.WRITE) as session:
        exceptions = lifetime_exception.add_exception(dids=dids, account=internal_account, pattern=pattern, comments=comments, expires_at=expires_at, session=session)

    for key in exceptions:
        if key == 'exceptions':
            for reqid in exceptions[key]:
                for did in exceptions[key][reqid]:
                    did['scope'] = did['scope'].external
                    did['did_type'] = did['did_type'].name
        else:
            for did in exceptions[key]:
                did['scope'] = did['scope'].external
                did['did_type'] = did['did_type'].name

    return exceptions


def update_exception(
    exception_id: str,
    state: 'LifetimeExceptionsState',
    issuer: str,
    vo: str = DEFAULT_VO,
) -> None:
    """
    Update exceptions state to Lifetime Model.

    :param id:         The id of the exception.
    :param state:      The states to filter.
    :param issuer:     The issuer account.
    :param vo:         The VO to act on.
    """
    kwargs = {'exception_id': exception_id, 'vo': vo}

    with db_session(DatabaseOperationType.WRITE) as session:
        auth_result = permission.has_permission(issuer=issuer, vo=vo, action='update_lifetime_exceptions', kwargs=kwargs, session=session)
        if not auth_result.allowed:
            raise exception.AccessDenied('Account %s can not update lifetime exceptions. %s' % (issuer, auth_result.message))
        return lifetime_exception.update_exception(exception_id=exception_id, state=state, session=session)
